
// $Id: Console.cpp,v 1.11 2001-04-09 22:06:37+02 damir Exp damir $

#include "stdafx.h"
#include <stdlib.h>
#include "Console.hpp"
#include "Errors.hpp"
#include "Auxiliary.hpp"


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

Direction _Scroll_Direction = dNoDirection;
Console   *_Console = NULL;


/*
* Console::Console(int numViewports, Environment *env)
* Desc: Creates timer, screen, viewports and a commandpanel and arranges viewports
* In:   numViewports is the number of viewports on screen, initializes video
*       env is the pointer passed to viewports
*/
Console::Console(int numViewports, EnvironmentClient *env)
{	
	// ignore numViewports for now
	
	_Console = this;
	drag = false;
	startDragPoint = TUI_Point(0, 0);
	autoScrollEnabled = false;
	
	OnWorldMouseDown = NULL;
	
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
		throw EErrorInitVideo();
	
	screen = SDL_SetVideoMode(
		CONSOLE_SCREEN_WIDTH, 
		CONSOLE_SCREEN_HEIGHT, 
		CONSOLE_BIT_DEPTH, 
		CONSOLE_VIDEO_FLAGS
		);
	
	if (screen == NULL)
		throw EErrorInitVideo();
	
	this->numViewports = numViewports;
	environment = env;	
	selected = new List();
	preselected =(List **) malloc(10*sizeof(List *));
	for (int i = 0; i < 10; i++)
		preselected[i] = new List();
	
	waypoints = new List();
	areasOfDanger = new List();

	activeViewport = 0;	
	ArrangeViewports(numViewports);
	
	
	scrollSpeed = SCROLL_SPEED; // default
	scrollStep = SCROLL_STEP;
	
	shiftState = 0;
	displayWaypoints = false;
	groupID = -1;
	isBeingScrolled = false;
	scrollStepsLeft = 0;
	currentDirection = dNoDirection;
	
	SDL_SetEventFilter(&InterceptEvents);
}


/*
* Console::~Console()
* Main destructor, destroys screen, viewports, timers...
*/
Console::~Console()
{
	for (int i = 0; i < numViewports; i++)
		delete viewports[i];
	free(viewports);
	delete commandPanel;
	delete selected;
	for (i = 0; i < 10; i++)
		delete preselected[i];
	free(preselected);
	
	waypoints->deleteAll(List::POINT);
	delete waypoints;
	areasOfDanger->deleteAll(List::AOD);
	delete areasOfDanger;
	environment = NULL;
	
	SDL_FreeSurface(screen);
	SDL_SetTimer(0, NULL);
	
	SDL_Quit();		
}


/*
* void Console::ArrangeViewports(int numViewports, Viewports **viewports)
* Desc: Arranges viewports according to number of viewports
* In:		numViewports is the number of viewports
*				viewports is the pointer to array of viewports
*/
void Console::ArrangeViewports(int numViewports)
{
	const int   VIEWPORT_BORDER = 3;
	const float VIEWPORT_CP_RATIO =(float) 1; 
	const float VCP =(float)0.7;
	int         i = 0;
	int         vH, vV;
	
	viewports = (Viewport**)(malloc(numViewports * sizeof(Viewport*)));
	commandPanel = new CommandPanel(this);
	
	// just for one, two and four viewports - not nice
	switch (numViewports)
	{
		case 1:
			viewports[0] = new Viewport(
				this, 
				screen->format, 
				(int)((screen->w - 2*VIEWPORT_BORDER) * VIEWPORT_CP_RATIO),
				(int)((screen->h - 2*VIEWPORT_BORDER) * VIEWPORT_CP_RATIO)
				);
			viewports[0]->top = VIEWPORT_BORDER;
			viewports[0]->left = VIEWPORT_BORDER;
			viewports[0]->Move(0, 0, GetScreen());
			break;
		case 2:				
			viewports[0] = new Viewport(
				this, 
				screen->format, 
				(int)((screen->w - 2*VIEWPORT_BORDER) * VCP),
				(int)((screen->h - 2*VIEWPORT_BORDER))
				);
			viewports[0]->top = VIEWPORT_BORDER;
			viewports[0]->left = VIEWPORT_BORDER;
			viewports[0]->Move(0, 0, GetScreen());
			
			viewports[1] = new Viewport(
				this, 
				screen->format, 
				(int)((screen->w - 3*VIEWPORT_BORDER) *(1 - VCP)),
				(int)((screen->h - 2*VIEWPORT_BORDER) *(1 - VCP))
				);
			viewports[1]->left = 2*VIEWPORT_BORDER + viewports[0]->width;
			viewports[1]->top = VIEWPORT_BORDER;
			viewports[1]->Move(0, 0, GetScreen());

			commandPanel->left = viewports[1]->left;
			commandPanel->top = viewports[1]->top + viewports[1]->height + VIEWPORT_BORDER;
			commandPanel->width = viewports[1]->width;
			commandPanel->height = viewports[0]->height - viewports[1]->height - VIEWPORT_BORDER;
			break;
		case 4:
			vV =(int)((screen->h - 3*VIEWPORT_BORDER) *(VIEWPORT_CP_RATIO /(float)2));
			vH =(int)((screen->w - 3*VIEWPORT_BORDER) *(VIEWPORT_CP_RATIO /(float)2));
			for (i = 0; i < numViewports; i++)
				viewports[i] = new Viewport(this, screen->format, vH, vV);												
			
			viewports[0]->left = VIEWPORT_BORDER;
			viewports[0]->top = VIEWPORT_BORDER;
			
			viewports[1]->left = 2*VIEWPORT_BORDER + vH;
			viewports[1]->top = VIEWPORT_BORDER;
			
			viewports[2]->left = VIEWPORT_BORDER;
			viewports[2]->top = 2*VIEWPORT_BORDER + vV;
			
			viewports[3]->left = 2*VIEWPORT_BORDER + vH;
			viewports[3]->top = 2*VIEWPORT_BORDER + vV;		
			
			for (i = 0; i < numViewports; i++)
				viewports[i]->Move(0, 0, GetScreen());			
			break;
	}
	SetActiveViewport(viewports[0]);
}


/*
* Direction Console::GetScrollingDirection(int x, int y)
* Desc: Gets the scrolling direction if the border of the screen was reached
* In:   x, y are the coordinates of the mouse
*				sx, sy, w, h are the bounds of the requested rect
* Out:  which direction to move the window to
*/ 
Direction	Console::GetScrollingDirection(int x, int y, int sx, int sy, int w, int h)
{
int vfac = (int)(SCROLL_DIAGONAL_FACTOR * h);
int hfac = (int)(SCROLL_DIAGONAL_FACTOR * w);
	
	if (x < sx + SCROLL_BORDER)			//mouse is at the left border
	{
		// upper border
		if (y < sy + vfac)
			return dUpLeft;
		else if (y > h - vfac)
			return dDownLeft;
		else 
			return dLeft;
	}
	 else if (x > w - SCROLL_BORDER)		//mouse is at the right border
	{
		// lower border
		if (y < sy + vfac)
			return dUpRight;
		else if (y > h - vfac)
			return dDownRight;
		else 
			return dRight;
	}
	 else if (y > h - SCROLL_BORDER)		//mouse is at the bottom border
	{
		if (x < sx + hfac)
			return dDownLeft;
		else if (x > w - hfac)
			return dDownRight;
		else 
			return dDown;
	}
	 else if (y < sy + SCROLL_BORDER)
	{
		if (x < sx + hfac)
			return dUpLeft;
		else if (x > w - hfac)
			return dUpRight;
		else 
			return dUp;
	}
	
	return dNoDirection;
}
	

/*
* void Console::WorldMouseDown(Viewport *sender, float x, float y, Uint8 button)
* Desc: all  viewports call this method after they've handled it. This way we can
*				set up a centralised response
*/
void Console::WorldMouseDown(Viewport *sender, float x, float y, Uint8 button)
{
	SoldierClient *soldier, *tSoldier;
	bool forwardEvent = true;

	//bool					s = false;		
	soldier = environment->GetSoldierAt(x, y); // only works for one soldier
	
	if (soldier != NULL){
		if (!shiftState){
			tSoldier = (SoldierClient*)selected->getFirst();
			while (tSoldier){
				tSoldier->selected = false;
				if (tSoldier->groupIndex == 0)
					tSoldier->groupIndex = -1;

				selected->remove();
				tSoldier = (SoldierClient*)selected->getNext();
			}
			groupID = -1;
		}

		
		if (soldier->isFriendly()){
			selected->add(soldier);
			soldier->selected = true;
						
			DefineGroup(0);
			RefreshOverlays();
			groupID = 0;	
			forwardEvent = false;
		}	

		if (commandPanel->OnChangeGroup)
			commandPanel->OnChangeGroup(0, true);
	} else if (commandPanel->GetButtonEnabled() == cpNone){
		tSoldier = (SoldierClient*)selected->getFirst();
		while (tSoldier){
			tSoldier->selected = false;
			if (tSoldier->groupIndex == 0)
				tSoldier->groupIndex = -1;

			selected->remove();
			tSoldier = (SoldierClient*)selected->getNext();
		}

		groupID = -1;
		if (commandPanel->OnChangeGroup)
			commandPanel->OnChangeGroup(0, true);				
	}		
	if (OnWorldMouseDown != NULL && forwardEvent)
			OnWorldMouseDown(x, y, button);
	
}


/*
* void Console::WorldMouseUp(Viewport *sender, float x, float y, Uint8 button)
* Desc: called by viewport on mouse up event, coordinates already transformed,
*       to reflect the map
* In:   Viewport is the calling viewport
*       x, y are the coordinates on the map
*       button is the button pressed
*/
void Console::WorldMouseUp(Viewport *sender, float x, float y, Uint8 button)
{
	SoldierClient *soldier;

	if (x != sender->GetStartX() && y != sender->GetStartY())
	{
		environment->SetAllSoldiersSelected(false);
		soldier =(SoldierClient*)(environment->soldiers->getFirst());
		
		while (soldier != NULL)
		{
			if (soldier->GetX() > MIN(sender->GetStartX(), x) && 
				soldier->GetX() < MAX(x, sender->GetStartX()) &&
				soldier->GetY() > MIN(sender->GetStartY(), y) && 
				soldier->GetY() < MAX(y, sender->GetStartY()))
			{
				if (soldier->isFriendly())
				{
					selected->add(soldier);
					groupID = 0;
					soldier->selected = true;
					soldier->groupIndex = 0;					
				}
			}
			
			soldier =(SoldierClient*)(environment->soldiers->getNext());
		} // while

		DefineGroup(0);
		RefreshOverlays();
		if (commandPanel->OnChangeGroup)
			commandPanel->OnChangeGroup(0, true);
	} // if	
}


/*
* void Console::RefreshOverlays()
* Desc: Calls refresh overlays for all viewports
*/
void Console::RefreshOverlays()
{
	int i = 0;
	
	for (i = 0; i < numViewports; i++)
		viewports[i]->RefreshOverlay(screen);
}


/*
* void Console::DefineGroup(int n)
* Desc: Stores currently selected soldiers as a predefined group for
*       easy futire reference.
* In:   n is the index of the predefined group
*/
void Console::DefineGroup(int n) 
{
	int group = 0;

	while (preselected[n]->getFirst())
		preselected[n]->remove();
	
	SoldierClient *tSoldier =(SoldierClient *) selected->getFirst();
	while (tSoldier) 
	{
		group = 0;
		while (group < 10)	//check all groups
		{
			SoldierClient *tmp = (SoldierClient *) preselected[group]->getFirst();			
			while (tmp && tmp != tSoldier)
				tmp = (SoldierClient *) preselected[group]->getNext();

			if (tmp)
			{
				preselected[group]->remove();
				break;
			}

			group++;
		}		//while

		preselected[n]->add(tSoldier);
		tSoldier->groupIndex = n;
		tSoldier =(SoldierClient *) selected->getNext();
	}
}


/*
* void Console::SelectGroup(int n)
* Desc: Selects soldiers form a predefined group.
* In:   n is the index of the predefined group
*/
void Console::SelectGroup(int n) 
{
	SoldierClient *tSoldier;
	while (tSoldier =(SoldierClient *) selected->getFirst()) 
	{
		tSoldier->selected = false;
		selected->remove();
	}
	
	tSoldier =(SoldierClient *) preselected[n]->getFirst();
	while (tSoldier) 
	{
		tSoldier->selected = true;
		selected->add(tSoldier);
		tSoldier =(SoldierClient *) preselected[n]->getNext();
	}
	groupID = n;
}


/*
* SDL_Surface Console::*GetScreen()
* Desc: Read the private property screen of the console
* Out:  Reference to screen
*/ 
SDL_Surface *Console::GetScreen()
{
	return screen;
}


/*
* EnvironmentClient *GetEnvironment(){
* Desc: returns the private environment
* Out:	Pointer to environment
*/
EnvironmentClient* Console::GetEnvironment(){
	return environment;
}


/*
* void Console::SetActiveViewport(int index)
* Desc: Sets the active viewport(also sets viewport's active member) if 
*       index is within bounds
* In:   index is the index of the new actve viewport 
*/
void Console::SetActiveViewport(int index)
{
	if (index >= 0 && index < numViewports)
	{
		if (activeViewport != NO_ACTIVE_VIEWPORT)
			viewports[activeViewport]->active = false;
		
		activeViewport = index;
		viewports[activeViewport]->active = true;
	}
}
	

/*
* void Console::SetActiveViewport(Viewport *viewport)
* Desc: Sets the active viewport(also sets viewport's active member) if 
*       pointer to viewport is legal
* In:   pointer of the new actve viewport 
*/
void Console::SetActiveViewport(Viewport *viewport)
{
	int i = 0;
	
	while (i < numViewports && viewport != viewports[i])
		i++;
	
	if (i < numViewports)
		SetActiveViewport(i);	
}

	
/*
* int Console::GetActiveViewport()
* Desc: Returns currently active viewport
* Out:	Currently active viewport
*/ 
int Console::GetActiveViewport()
{
	return activeViewport;
}


/*
* int Console::GetNumViewports()
* Desc: returns the private numViewports member
* Out:	number of viewports
*/ 
int Console::GetNumViewports()
{
	return numViewports;
}

	
/*
* int	InterceptEvents(const SDL_Event *event)
* Desc: main event handler. All mouse events are passed to all viewports until one handles it.
*       All key events are passed to command panel
*/
int	InterceptEvents(const SDL_Event *event)
{		
	int    i = 0;
	if (event->type == SDL_MOUSEMOTION)
	{						
		// motion events are initiated at borders even if cursor is not moved

		if (!_Console->isBeingScrolled || !_Console->autoScrollEnabled){		//auto scroll has priority
			_Scroll_Direction = _Console->GetScrollingDirection(event->motion.x, event->motion.y, 0, 0, CONSOLE_SCREEN_WIDTH, CONSOLE_SCREEN_HEIGHT); // set tmp vars
			
			if (_Scroll_Direction != dNoDirection && _Console->currentDirection != _Scroll_Direction)
				SDL_SetTimer(1, &OnScrollTimer);  // immediate call
			else if (_Scroll_Direction == dNoDirection) // scroll_direction
				SDL_SetTimer(0, NULL);            // detach timer
			
			_Console->currentDirection = _Scroll_Direction;
		}
			
		if (_Console->GetActiveViewport() != NO_ACTIVE_VIEWPORT)
		{
			// handle mousemotion for active viewport
			float rX = 0, rY = 0;
			if (_Console->viewports[_Console->GetActiveViewport()]->MouseMotion(
				_Console->GetScreen(),
				event->motion.x,
				event->motion.y,
				&rX,
				&rY
				))
				SDL_Flip(_Console->GetScreen());
		} // if	
		return (0);
	} // mousemotion
	
	if (event->type == SDL_MOUSEBUTTONDOWN)
	{			
		float rX = 0, rY = 0;
		for (i = 0; i < _Console->GetNumViewports(); i++)
		{
			// check to see which viewport has buttondown
			if (_Console->viewports[i]->MouseButtonDown(
				_Console->GetScreen(),
				event->button.x,
				event->button.y,
				&rX,																				
				&rY,
				event->button.button))
			{
				_Console->WorldMouseDown(_Console->viewports[i], rX, rY, event->button.button);
				_Console->SetActiveViewport(i);				
				
				_Console->RefreshOverlays();
				SDL_Flip(_Console->GetScreen());
				break;
			} // if
		}		
	} // mousedown
	
	if (event->type == SDL_MOUSEBUTTONUP)
	{
		if (_Console->GetActiveViewport() != NO_ACTIVE_VIEWPORT)
		{
			// forces mouseup in same viewport as mousedown
			float rX = 0, rY = 0;
			_Console->viewports[_Console->GetActiveViewport()]->MouseButtonUp(
				_Console->GetScreen(),
				event->button.x, 
				event->button.y,
				&rX,
				&rY,
				event->button.button
				);					
			_Console->WorldMouseUp(
				_Console->viewports[_Console->GetActiveViewport()], 
				rX,
				rY,
				event->button.button
				);
			_Console->RefreshOverlays();
			SDL_Flip(_Console->GetScreen());
		} // if no_active_vieport
	} // mouseup
	
	if (event->type == SDL_KEYDOWN)
	{
		switch (event->key.keysym.sym)
		{
			case SDLK_KP_PLUS:
				_Console->viewports[_Console->GetActiveViewport()]
					->SetZoom(_Console->viewports[_Console->GetActiveViewport()]->GetZoom() *
					CONSOLE_ZOOM_STEP, _Console->GetScreen());
				SDL_Flip(_Console->GetScreen());
				return (0);
				break;
			case SDLK_KP_MINUS:
				_Console->viewports[_Console->GetActiveViewport()]
					->SetZoom(_Console->viewports[_Console->GetActiveViewport()]->GetZoom() /
					CONSOLE_ZOOM_STEP, _Console->GetScreen());
				SDL_Flip(_Console->GetScreen());
				return (0);
				break;
			case SDLK_ESCAPE:
				exit(0);			
				break;	
			case SDLK_RSHIFT:
			case SDLK_LSHIFT:
				_Console->shiftState++;
				break;			
			case SDLK_1:
			case SDLK_2:
			case SDLK_3:
			case SDLK_4:
			case SDLK_5:
			case SDLK_6:
			case SDLK_7:
			case SDLK_8:
			case SDLK_9:
				if ((event->key.keysym.mod == KMOD_LCTRL) ||
					(event->key.keysym.mod == KMOD_RCTRL)){
					_Console->DefineGroup(event->key.keysym.sym - SDLK_0);
					_Console->SelectGroup(event->key.keysym.sym - SDLK_0);
					_Console->RefreshOverlays();
				}else if ((event->key.keysym.mod == KMOD_LSHIFT) ||
					(event->key.keysym.mod == KMOD_RSHIFT)){
					/*SoldierClient *soldier = (SoldierClient*)_Console->selected->getFirst();

					while (soldier){
						soldier->SetAgr((float)(event->key.keysym.sym - SDLK_0)/(float)(10));
						soldier = (SoldierClient*)_Console->selected->getNext();
					}
					
					_Console->RefreshOverlays();*/					
				} else {
					_Console->SelectGroup(event->key.keysym.sym - SDLK_0);
					_Console->RefreshOverlays();}
			case SDLK_F2:
				_Console->autoScrollEnabled = !_Console->autoScrollEnabled;
			default:
				if (_Console->commandPanel->ProcessEvent(event, _Console->GetScreen()))
					return (0);
		} // switch
	} // keydown
	
	if (event->type == SDL_KEYUP)
	{
		switch (event->key.keysym.sym)
		{
			case SDLK_RSHIFT:
			case SDLK_LSHIFT:
				_Console->shiftState--;
				break;
			default:
				if (_Console->commandPanel->ProcessEvent(event, _Console->GetScreen()))
					return 0;
		}
	}
	return (1);
}

	
/*
* Uint32 OnScrollTimer(Uint32 interval)
* Desc: depending on dir starts calling Move method of the active viewport - sets the timer		
*/
Uint32 OnScrollTimer(Uint32 interval)
{
	static Viewport* active;

	if (_Console->isBeingScrolled){			//one step of auto scroll
		_Console->scrollStepsLeft--;

		if (_Console->scrollStepsLeft == 0){		//end auto scroll
			SDL_SetTimer(0, NULL);
			_Console->isBeingScrolled = false;
		}
	}

	active = _Console->viewports[/*_Console->GetActiveViewport()*/0];

	active->Move(
		_Console->currentDirection,
		_Console->scrollStep,
		_Console->GetScreen()
		);		

	//SDL_Flip(_Console->GetScreen());
	SDL_UpdateRect(
		_Console->GetScreen(),
		active->left, 
		active->top, 
		active->width,
		active->height
	);
	
	return _Console->scrollSpeed;		
}
