
// $Id: List.cpp,v 1.11 2001/04/15 15:06:31 mitja Exp $

/* List class implementation
 *
 * Created by: Damir Arh
 * Created on: 19.11.2000
 */

#include "stdafx.h"
#include <stdlib.h>
#include "List.hpp"

#ifdef __SERVER__
#include "Soldier.hpp"
#include "Group.hpp"
#include "packages.hpp"
#endif

#ifdef __CLIENT__
#include "SoldierClient.hpp"
#include "BulletClient.hpp"
#include "TuiGraphics.hpp"
#include "AreaOfDanger.hpp"
#endif

#include <crtdbg.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif



/* List()
 *
 * Creates an empty list.
 */
List::List() 
{
	first = NULL;
	current = NULL;
	next = NULL;
	count = 0;
}


/* ~List()
 *
 * Destroys the list by freeing all list nodes but keeping the items in those
 * nodes(i.e. memory has to be freed manually but the items are still available
 * after the list gets destroyed). You can use the deleteAll() method before
 * destroying the list if you don't want to free the items manually.
 */
List::~List() 
{
	ListNode *temp;
	// delete the first node while available
	while (first != NULL) 
	{
		temp = first;
		first = first->next;
		delete temp;
	}
}


/* getFirst()
 *
 * Result:
 *    void * : pointer to a stored variable or NULL if list is empty
 *
 * Returns the item in the first list node and sets internal current pointer
 * to the first node to enable the functionality of getNext() method. The
 * returned variable is not copied, so do not free it before removing it from
 * the list.
 */
void *List::getFirst() 
{
	current = first;
	if (current != NULL)
	{
		// is the list non-empty?
		next = current->next;
		return current->item;
	}
	else
		return NULL;
}


/* getNext()
 *
 * Result:
 *    void * : pointer to a tored variable or NULL if there's no next item
 *
 * Returns the next item in the list in respect to the last returned variable.
 * Updates the internal current counter accordingly, of course. Returns NULL
 * if there is no next item or if no item has been yet read by using getFirst().
 * Again the returned variable has to be removed from the list before freeing
 * it.
 */
void *List::getNext() 
{
	if (next != NULL)
	{
		// are there any more nodes?
		current = next;       // set return node
		next = current->next; // set next node for return
		return current->item;
	}
	current = NULL; // no node is returned
	return NULL;
}

/* add(void *newItem)
 *
 * Parameters:
 *    newItem : pointer to the variable to store in the list
 *
 * Inserts a new node which contains newItem at the beginning of the list.
 * The current node remains unchanged.
 */
void List::add(void *newItem) 
{
	ListNode *newNode = new ListNode(newItem); // store the variable in a node
	
	if (first != NULL)
	{
		// if list is not empty
		// update pointers which link the list
		first->prev = newNode;
		newNode->next = first;
	}
	first = newNode; // set pointer to the first node
	count++;
}


/* remove()
 *
 * Removes the current node from the list but doesn't free the stored item
 * which is still accessible even after this operation. The current node is
 * the last one which was returned by either getFirst() or getNext(). If there
 * is no current valid node the method does nothing. After the node is removed
 * additional removal without changing the current node with getFirst() or
 * getNext() will do nothing, though the getNext() method behave is this node
 * has not been removed.
 */
void List::remove() 
{
	if (current != NULL)
	{
		// if current node valid
		if (current->prev != NULL)               // if not first node
			current->prev->next = current->next; // previous node skips this one
		else
			first = current->next;               // new first node
		if (current->next != NULL)               // if not last node
			current->next->prev = current->prev; // next node skips this one
		delete current;
		current = NULL; // current node not valid any more
		count--;
	}
}


/* deleteAll()
 *
 * Empties the list by deleting all the nodes including the values stored in
 * them. Those values can not be referenced anymore after this operation. Use
 * it only when this list is the only remaining way of reaching them and you
 * don't need them any more.
 */
void List::deleteAll(tType type) 
{
	void *item;
	
	while (first != NULL) 
	{ // while list not empty
//		_RPTF1(_CRT_WARN, "bla %ld\n", type);
		item = getFirst();  // get first item
		remove();           // free node
		switch (type)       // free item
		{
#ifdef __SERVER__
		case SOLDIER:
			delete ((Soldier *) item);
			break;
		case GROUP:
			delete ((Group *) item);
			break;
		case BULLET:
			delete ((Bullet *) item);
			break;
		case VECTOR:
			delete ((Vector *) item);
			break;
		case INT:
			delete ((int *) item);
			break;
		case VERTEX:
			delete ((VGVertex *) item);
			break;
		case AOD:
			delete ((AODpack *) item);
			break;
#endif
#ifdef __CLIENT__
		case CSOLDIER:
			delete ((SoldierClient *) item);
			break;
		case CBULLET:
			delete ((BulletClient *) item);
			break;
		case POINT:
			delete ((TUI_pointStruct *) item);
			break;
		case AOD:
			delete ((AreaOfDanger *) item);
			break;
#endif
		}
	}
	// remove references to non-existing nodes
	current = NULL;
	next = NULL;
	count = 0;
}

void List::empty() 
{
	void *item;
	
	while (first != NULL) 
	{ // while list not empty
		item = getFirst();  // get first item
		remove();           // free node
	}
	// remove references to non-existing nodes
	current = NULL;
	next = NULL;
	count = 0;
}

/* getNumItems()
 *
 * Result:
 *    int : number of items in the list
 *
 * Returns the current number of items in the list.
 */
int List::getNumItems() 
{
	return count;
}

void List::bookmark(int index)
{
	bookmarks[index] = current;
}

void *List::goTo(int index)
{
	current = bookmarks[index];
	if (current != NULL)
	{
		next = current->next;
		return current->item;
	}
	else
	{
		next = NULL;
		return NULL;
	}
}

void List::addList(List *list)
{
	List *interList = new List;
	void *interItem;

	interItem = list->getFirst();
	while (interItem != NULL)
	{
		interList->add(interItem);
		interItem = list->getNext();
	}
	interItem = interList->getFirst();
	while (interItem != NULL)
	{
		add(interItem);
		interItem = interList->getNext();
	}
	delete interList;
	delete list;
}

void List::merge(List *list)
{
	ListNode *temp;
	void *item;
	bool done;

	item = list->getFirst();
	while (item)
	{
		temp = first;
		done = false;
		while ((temp) && (!done))
		{
			if (temp->item != item)
				temp = temp->next;
			else
				done = true;
		}
		if (!done)
			add(item);
		item = list->getNext();
	}
}


//
// doda na konec seznama - GREGOR
//
void List::addLast(void *item)
{
	ListNode * novi = new ListNode(item);
	ListNode * temp = first;
	
	if (temp == NULL) 
	{
		add(item);
		return;
	}

	while (temp->next != NULL) temp = temp->next;
	temp->next = novi;
	novi->prev = temp;

	count++;
}

// MITJA
bool List::contains (void *item)
{
	ListNode *temp=first;

	while (temp!=NULL)
	{
		if (temp->item==item)
			return true;
		temp=temp->next;
	}
	return false;
}