
// $Id: Soldier.hpp,v 1.16 2001/04/08 15:40:52 mitja Exp $

/* Soldier class
 *
 * This is the main class for all units in the game. It contains all the information about
 * the unit as well as methods for controlling the unit, getting information about it and
 * setting its state (only to be used by other classes of the Physics engine). The 
 * getCoords() / getNewCoords() pair returns relevant movement info to the environment class.
 *
 * The class will very probably be extended in the future with methods which will enable
 * to take control of the unit at a higher abstraction layer and will then call the low
 * level functions.
 *
 * Created by: Damir Arh
 * Created on: 17.11.2000
 */

#ifndef __SOLDIER__
#define __SOLDIER__

#include "Vector.hpp"
#include "Bullet.hpp"

class Environment;

class Soldier
{
public:
	// enumeration declaration
	enum tMovement { MOVEMENT_WALK, MOVEMENT_CROUCH, MOVEMENT_CRAWL };              // types of movement
	enum tShoot { SHOOT_AIM, SHOOT_SNAP, SHOOT_AUTO, SHOOT_NONE };                  // types of shooting
	enum tAngleDifference { DIFFERENCE_SHOOT_UNIMPEDED, DIFFERENCE_SHOOT_IMPEDED }; // types of maximum allowed differences between angles of movement and view

private:
	friend class Environment;

	// unique IDs to identify the soldier and its AI
	int soldierID;
	int aiID;

	// soldier internal values (not to be changed for now, default values for all soldiers)
	float radius;               // radius of the soldier (0.3 m)
	float maxSpeed;             // maximum movement speed (6 m/s)
	float backSpeedFactor;      // factor for maximum speed of moving backwards (0.6)
	float *speedFactor;         // array with factors for maximum speed per movement type (1, 0.6, 0.15)
	float *movementChangeDelay; // array with delays for changing movement type (0 s, 0.2 s, 0.8 s; 0,2 s, 0 s, 0.7 s; 1.5 s, 1.2 s, 0 s)
	float *hitRadiusFactor;     // factors by which radius is multiplied when calculating probability of being hit in relation to movement type (1, 0.6, 0.2)
	float accuracyTangens;      // tangens of the soldier accuracy triangle (0.003)
	float *shootDelay;          // array with delays for different types of shooting (3 s, 1 s, 0.1 s, 0 s)
	float *accuracyFactor;      // array with accuracy tangens multipliers for different types of shooting (0.5, 1, 3, 0)
	float rotationMoveSpeed;    // speed of movement rotation (2*PI rad/s)
	float rotationViewSpeed;    // speed of view rotation (4*PI rad/s)
	float acceleration;         // movement acceleration (3 m/s2)
	float decceleration;        // movement decceleration (5 m/s2)
	float leftShootLimit;       // angle to the left to which soldier can shoot unimpeded (0.125*PI rad) | Because y cordinates increase in the opposite direction from
	float rightShootLimit;      // angle to the right to which soldier can shoot unimpeded (0.5*PI rad)  | what was expected, left and right are reversed.
	float timeLastTurn;         // time of previous update
	float tick;                 // time between updates as deduced from the last two
	float *fieldOfViewLimits;   // minimum and maximum field of view value (PI/3 rad, 0.9*PI rad)

	// current soldier's state (updated every turn)
	Vector coords;      // soldier position [m, m]
	Vector oldCoords;   // soldier previous position from [m, m]
	float moveAngle;    // angle of movement [rad]
	float viewAngle;    // angle of view (and shooting) [rad]
	float speed;        // speed of movement [m/s]
	tMovement moveType; // type of movement

	// last soldier orders (set by low level ordering methods)
	float desiredViewAngle;               // ordered angle of view (and shooting) [rad]
	float desiredViewAngleRel;            // ordered relative angle of view (and shooting): positive value means right, negative value means left [rad]
	Vector desiredViewCoords;             // ordered coordinates for soldier to look towards [m, m]
	float desiredMoveAngle;               // ordered angle of movement [rad]
	float desiredSpeed;                   // ordered speed [m/s]
	tMovement desiredMoveType;            // ordered type of movement
	float desiredMaxLeftAngleDifference;  // ordered maximum difference between angle of movement and angle of view when looking left [rad]  | Because y cordinates increase in the opposite direction from
	float desiredMaxRightAngleDifference; // ordered maximum difference between angle of movement and angle of view when looking right [rad] | what was expected, left and right are reversed.
	bool desiredCanReverse;               // permission to reverse direction of movement
	float timeMoveChanged;                // time of movement type order [s]
	tShoot desiredShootType;              // ordered type of shooting
	float timeLastShoot;                  // time of shooting order [s]

	// soldier states to be updated by other Physics engine classes
	int health;        // soldier health (0-255)
	float shotAtAngle; // angle of the last missed shot [rad]
	int agression;

public:
	// constructors and destructors
	Soldier(int soldierID, int aiID, Vector coords, float angle);
	~Soldier();

	// ordering methods
	void setViewAngle(float angle);                                         // set desired angle of view
	void setViewAngleRel(float angle);                                      // set desired angle of view relatively to angle of movement: positive value means right, negative value means left
	void setViewAnglePoint(Vector coords);                                  // set coordinates soldier is to look towards
	void setMoveAngle(float angle);                                         // set desired angle of movement
	void setSpeed(float speed);                                             // set desired speed
	void setMoveType(tMovement type);                                       // set desired movement type
	void setMaxAngleDifference (float difference, bool reverse);            // set desired maximum difference between angle of movement and angle of view and whether soldier is permitted to reverse direction of movement
	void setMaxAngleDifference (tAngleDifference difference, bool reverse); // set desired maximum difference between angle of movement and angle of view and whether soldier is permitted to reverse direction of movement
	void setMoveVector(Vector moveVector);                                  // set desired speed and angle of movement with given vector
	void setAgression(int agr);												// set soldiers agression
	void shoot(tShoot type);                                                // order a shot

	// methods to be called by Bullet class
	void setShotAtAngle(float angle);  // angle at which bullet missed the soldier
	void setIsShot(float hitAccuracy); // soldier was hit

	// methods for immediate soldier updating
	Vector getCoords();              // get current coordinates
	Vector getOldCoords();           // get previous coordinates 
	Vector getNewCoords(float time); // get desired coordinates
	Bullet *getBullet();             // get the bullet shot in the last turn if any
	void stop(Vector coords);        // tells the soldier to stop at the given coordinates immediately

	// methods to obtain info about the soldier
	float getRadius();       // get soldier radius
	float getHitRadius();    // get soldier radius for calculating whether he's hit
	int getHealth();         // get soldier health
	float getMoveAngle();    // get soldier angle of movement
	float getViewAngle();    // get soldier angle of view
	float getSpeed();        // get soldier speed
	float getFieldOfView();  // get soldier field of view
	tMovement getMoveType(); // get soldier type of movement
	int getID();             // get soldier unique ID
	int getAiID();           // get AI ID of soldier owner
	Vector getMoveVector();  // get vector of movement
	int getAgression();      // get soldiers agression

	// methods to obtain additional info about soldier
	float calculateMaxSpeed(tMovement type); // calculate maximum speed for current difference between angle of movement and angle of view and given movement type
	float calculateAccuracyTangens();        // calculate current tangens of the soldier accuracy triangle, return negative value if shooting impossible
};

#endif // __SOLDIER__
