
// $Id: Viewport.cpp,v 1.7 2001/04/08 20:59:44 mitja Exp $

#include "stdafx.h"
#include <math.h>
#include "Viewport.hpp"
#include "Errors.hpp"
#include "SoldierClient.hpp"
#include "BulletClient.hpp"
#include "TUIGraphics.hpp"
#include "Auxiliary.hpp"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*
* Uint32 MapColor(MyColor color, SDL_PixelFormat *fmt)
* Desc: maps a constant in an enum type to a specified color using pixel format
* In:   color is the constant of the color
*       fmt is the pixel type of the dest surace
*/
Uint32 MapColor(MyColor color, SDL_PixelFormat *fmt)
{
	Uint8 r, g, b;
	
	switch (color)
	{
		case clBlack: 
			r = 0;
			b = 0;
			g = 0;
			break;
		case clBlue: 
			r = 0;
			b = 220;
			g = 0;
			break;
		case clYellow: 
			r = 255;
			b = 0;
			g = 255;
			break;
		case clWhite: 
			r = 255;
			b = 255;
			g = 255;
			break;
		case clRed: 
			r = 255;
			b = 0;
			g = 0;
			break;
		case clGreen: 
			r = 0;
			b = 0;
			g = 255;
			break;
		case clCyan: 
			r = 128;
			b = 255;
			g = 255;
		case clBackground: 
			r = 235,
			b = 235;
			g = 235;
			break;
		case clSeparatorLines: 
			r = 220;
			b = 220;
			g = 220;
			break;
		case clFilled: 
			r = 100;
			b = 160;
			g = 100;
			break;	
	} // switch

	return SDL_MapRGB(fmt, r, g, b);
}


/*
* Viewport::Viewport(Console *console, SDL_PixelFormat *fmt, int w, int h)
* Desc: Main constructor, creates new offscreen surface, with fmt format and sets
*       environment, and bounds of a viewport
* In:   console is the pointer to console object
*       fmt is the pixel format of the screen used for offscreening
*       w, h width and height			
*/
Viewport::Viewport(Console *console, SDL_PixelFormat *fmt, int w, int h)
{
	if (console->GetEnvironment() == NULL) 
		throw EErrorViewportNoEnvironment();
	
	this->console = console;
	offScreen = SDL_CreateRGBSurface(0, w, h, fmt->BitsPerPixel, fmt->Rmask, fmt->Bmask,
		fmt->Gmask, fmt->Amask); 		
	
	startX = 0;
	startY = 0;
	endX = 0;
	endY = 0;
	width = w;
	height = h;
	top = 0;
	left = 0;
	posX = 0;
	posY = 0;
	fZoom = 1;
	drag = false;
}


/*
*	Viewport::~Viewport()
* Desc:	Main destructor, disposes of offscreen and sets console to NULL
*/		
Viewport::~Viewport()
{
	SDL_FreeSurface(offScreen);
	console = NULL;
}


/*
* float Viewport::GetZoom()
* Desc: getting the private fZoom
* Out:	fZoom
*/
float	Viewport::GetZoom()
{
	return fZoom;
}


/*
* float Viewport::GetStartX()
* Desc: get the starting x
* Out:  private startX member 
*/
float Viewport::GetStartX()
{
	return startX;
}


/*
* float Viewport::GetStartY()
* Desc: get the starting y
* Out:  private startY member 
*/	
float Viewport::GetStartY()
{
	return startY;
}
	

/*
* void Viewport::SetZoom(float zoom, SDL_Surface *screen)
* Desc: setting the private fZoom
* In:   new fZoom
*       screen is the destination surface
*/
void Viewport::SetZoom(float zoom, SDL_Surface *screen)
{
	fZoom = zoom;
	
	if (screen != NULL)
		Move(posX, posY, screen);
}
	

/*
* bool Viewport::MouseButtonDown(SDL_Surface *screen, int x, int y, float *rX, float *rY, Uint8 Button)
* Desc:	Converts the received coordinates to absolute world coordinates and returns
*       true if x, y are really inside viewport
* In:   x, y mouse coordinates(absolute to screen)
*       Button is the button pressed
*       screen is needed for drag rect output
* Out:  rX, rY are the absolute world coordinates
*/
bool Viewport::MouseButtonDown(SDL_Surface *screen, int x, int y, float *rX, float *rY,
							   Uint8 Button)
{
	if (x > left && y > top && x < left + width && y < top + height)
	{
		*rX = posX +(x - left)/fZoom;
		*rY = posY +(y - top)/fZoom;
		
		drag = true;
		
		startX = *rX;
		startY = *rY;
		endX = *rX;
		startY = *rY;
		
		return true;
	}
	else
		return false;
}


/*
* bool Viewport::MouseButtonUp(SDL_Surface *screen, int x, int y, float *rX, float *rY, Uint8 Button)
* Desc:	Converts the received coordinates to absolute world coordinates and returns
*       true if x, y are really inside viewport
* In:   x, y mouse coordinates(absolute to screen)
*       Button is the button pressed
*       screen is needed for drag rect output
* Out:  rX, rY are the absolute world coordinates
*/
bool Viewport::MouseButtonUp(SDL_Surface *screen, int x, int y, float *rX, float *rY, Uint8 Button)
{
	if (x > left && y > top && x < left + width && y < top + height)
	{
		*rX = posX +(x - left)/fZoom;
		*rY = posY +(y - top)/fZoom;
		
		drag = false;
		
		DrawScreenRect(screen, startX, startY, *rX, *rY);
		
		return true;
	}
	else
		return false;
}


/*
* bool Viewport::MouseButtonUp(SDL_Surface *screen, int x, int y, float *rX, float *rY, Uint8 Button)
* Desc:	Converts the received coordinates to absolute world coordinates and returns
*       true if x, y are really inside viewport
* In:   x, y mouse coordinates(absolute to screen)
*       screen is needed for drag rect output
* Out:  rX, rY are the absolute world coordinates
*/
bool Viewport::MouseMotion(SDL_Surface *screen, int x, int y, float *rX, float *rY)
{
	if (x > left && y > top && x < left + width && y < top + height && drag)
	{
		if (drag)
			DrawScreenRect(screen, startX, startY, endX, endY); // delete the last
		
		*rX = posX +(x - left)/fZoom;
		*rY = posY +(y - top)/fZoom;
		
		if (drag)
			DrawScreenRect(screen, startX, startY, *rX, *rY); // draw a new one
		
		endX = *rX;
		endY = *rY;
		
		return true;
	}
	else
		return false;
}


/*
* void Viewport::DrawBase(float oldX, float oldY, float newX, float newY)
* Desc:	Incorporates the logic for blitting the unchanged part of environment and redrawing the 
*       newly exposed and therefore blank part of the viewport. Refreshes the blank part with
*       fresh textures
* In:   oldX, oldY are the old posX and posY
*       newX, newY are the new
*/
void Viewport::DrawBase(float oldX, float oldY, float newX, float newY)
{
	int      numX, numY, startX, startY;
	int      i, j;
	float    side;
	SDL_Rect rect;
	
	side = console->GetEnvironment()->GetSquareBorder() * fZoom;
	
	startX = (int) newX / console->GetEnvironment()->GetSquareBorder();
	startY = (int) newY / console->GetEnvironment()->GetSquareBorder();
	
	numX = (int)(width / (int)side) +  2 + startX; 
	numY = (int)(height / (int)side) +  2 + startY;
	
	rect.x = 0;
	rect.y = 0;
	rect.w = width;
	rect.h = height;
	// resets bkg
	SDL_FillRect(offScreen, &rect, MapColor(clSeparatorLines, offScreen->format));
	
	for (i = startX; i <= MIN(numX, console->GetEnvironment()->GetMaxSX()- 1); i++)
	{
		for (j = startY; j <= MIN(numY, console->GetEnvironment()->GetMaxSY() - 1); j++)
		{
			rect.x =(int)((i - startX) * side - (int)((int) newX % console->GetEnvironment()->GetSquareBorder() * fZoom + 0.5) + 0.5);
			rect.y =(int)((j - startY) * side - (int)((int) newY % console->GetEnvironment()->GetSquareBorder() * fZoom + 0.5) + 0.5);
			//rect.w =(int)(side - 2 + 0.5); // draw terrain
			//rect.h =(int)(side - 2 + 0.5);
			rect.w =(int)(side + 1); // draw terrain
			rect.h =(int)(side + 1);
			
			if (console->GetEnvironment()->GetSquare(i, j)->passability == 0) 
				SDL_FillRect(offScreen, &rect, MapColor(clFilled, offScreen->format));
			else	
				SDL_FillRect(offScreen, &rect, MapColor(clBackground, offScreen->format));
		} // j
	} // i		
}


/*
* void Viewport::Move(float x, float y, Rects *rects; SDL_Surface *screen)
* Desc:	calls DrawBase with  the required parameters, then sets the new posX, posY and as last
*       calls the DrawOverlay. It also sets the dirty rects
* In:   x, y are the new position
*       screen is the destination screen
*/				
void Viewport::Move(float x, float y, SDL_Surface *screen)
{
	SDL_Rect src, dest;

	clipRect.x = float(left);
	clipRect.y = float(top);
	clipRect.w = float(width);
	clipRect.h = float(height);

	DrawBase(posX, posY, x, y);
	
	src.x = 0;
	src.y = 0;
	src.w = width;
	src.h = height;
	
	dest.x = left;
	dest.y = top;
	dest.w = width;
	dest.h = height;
	
	SDL_BlitSurface(offScreen, &src, screen, &dest);				
	
	posX = x;
	posY = y;
	
	DrawOverlay(screen);
	
	if (drag)
		DrawScreenRect(screen, startX, startY, endX, endY);
}


/*
* void Viewport::Move(Direction dir, int scrollSpeed, Rects *rects, SDL_Surface *screen)
* Desc: sets the new coordinates according to direction and then calls Move(x, y, Rects*, SDL_Surface).
*       Moving is stopped if the border is reached, any further movement is this direction 
*       is ignored
* In:   dir is the direction to move into
*       screen is the destination surface
*       scrollStep is the number of pixels to move
*/
void Viewport::Move(Direction dir, int scrollStep, SDL_Surface *screen)
{
	float tmpX = posX, tmpY = posY;
	int step =(int)(scrollStep / fZoom);


	if (dir == dUp || dir == dUpLeft || dir == dUpRight)
	{
		// up
		if (posY > scrollStep / fZoom)	
			tmpY = posY - step;
		else
			tmpY = 0;
	}
	
	if (dir == dDown || dir == dDownLeft || dir == dDownRight)
	{
		// down
		if (posY + height / fZoom + step <
			(console->GetEnvironment()->GetMaxSY() * console->GetEnvironment()->GetSquareBorder()))
		{
			tmpY = posY + step;
		}
		else
		{
			tmpY =(float)(console->GetEnvironment()->GetMaxSY() * console->GetEnvironment()->GetSquareBorder()) -
				height / fZoom;
		}
	}
	
	if (dir == dLeft || dir == dUpLeft || dir == dDownLeft)
	{
		// left
		if (posX > scrollStep / fZoom)	
			tmpX = posX - step;
		else
			tmpX = 0;
	}
	
	if (dir == dRight || dir == dDownRight || dir == dUpRight)
	{
		// right
		if (posX + width / fZoom + step <
			(console->GetEnvironment()->GetMaxSX() * console->GetEnvironment()->GetSquareBorder()))	
		{
			tmpX = posX + step;
		}
		else
		{
			tmpX =(float)(console->GetEnvironment()->GetMaxSX() * console->GetEnvironment()->GetSquareBorder()) -
				width / fZoom;
		}
	}
	
	if (tmpX < 0)
		tmpX = 0;
	
	if (tmpY < 0)
		tmpY = 0;
	
	Move(tmpX, tmpY, screen);
}


/*
* void Viewport::DrawOverlay()
* Desc: Draws all the overlay objects on offScreen using this viewport's
*       properties
*/
void Viewport::DrawOverlay(SDL_Surface *surface)
{
	const int WP_SIZE = 30;

	SoldierClient *soldier;
	BulletClient  *bullet;
	TUI_pointStruct *p;
	AreaOfDanger *area;
	
	soldier = (SoldierClient*)(console->GetEnvironment()->soldiers->getFirst()); // soldiers
	
	while (soldier != NULL) 
	{
		soldier->DrawSoldier(this, surface); 
		soldier =(SoldierClient*)(console->GetEnvironment()->soldiers->getNext());
	}		//while
	
	bullet =(BulletClient*)(console->GetEnvironment()->bullets->getFirst()); // bullets
	
	while (bullet != NULL)
	{
		bullet->DrawBullet(this, surface); 
		bullet =(BulletClient*)(console->GetEnvironment()->bullets->getNext());
	}			//while

	p = (TUI_pointStruct*)(console->waypoints->getFirst());

	while ((p != NULL) && (console->displayWaypoints))
	{

		if (PointInViewport(p->x, p->y))		
			DrawWaypoint(
				surface, 
				(int)((p->x - posX) * fZoom + (float)left), 
				(int)((p->y - posY) * fZoom + (float)top), 
				(int)(WP_SIZE * fZoom)
			);
		p = (TUI_pointStruct*)(console->waypoints->getNext());			
	}		//while
	
	area = (AreaOfDanger*)(console->areasOfDanger->getFirst());

	while (area != NULL) {		
		DrawAOD(
			surface, 
			(int)((area->x - posX) * fZoom + (float)left), 
			(int)((area->y - posY) * fZoom + (float)top), 
			(int)(area->r * fZoom)
		);
		
		area = (AreaOfDanger*)(console->areasOfDanger->getNext());			
	}

}


/* 
* void Viewport::RefreshOverlay(SDL_Surface *surface)
* Desc: Blits offScreen with surface and then calls drawOverlay procedure
* In:   surface is the destination surface
*/
void Viewport::RefreshOverlay(SDL_Surface *surface)
{
	SDL_Rect src, dest;

	clipRect.x = float(left);
	clipRect.y = float(top);
	clipRect.w = float(width);
	clipRect.h = float(height);

	src.x = 0;
	src.y = 0;
	src.w = width;
	src.h = height;
	
	dest.x = left;
	dest.y = top;
	dest.w = width;
	dest.h = height;
	
	SDL_BlitSurface(offScreen, &src, surface, &dest);
	
	DrawOverlay(surface);	
	
	if (drag)
		DrawScreenRect(surface, startX, startY, endX, endY);
}


/*
* bool Viewport::PointInViewport(float x, float y)
* Desc: Checks for point inside the viewport
* In:   x, y is the point to be checked
* Out:  true if inside viewport, false otherwise
*/
bool Viewport::PointInViewport(float x, float y)
{
	if ((x > posX && x < posX + width / fZoom) &&
		(y > posY && y < posY + height / fZoom))
	{
		return true;
	}
	else
	{
		return false;
	}
}


/*
* bool Viewport::RectInViewport(TUI_rectStruct r) 
* Desc: Checks for rect inside viewport
* In:   r is the rectangle to be checked
* Out:  VIEWPORT_RECT_OUTSIDE if outside, VIEWPORT_RECT_INSIDE if the
*       whole rect is inside or VIEWPORT_RECT_OVERLAPS if at least
*       a part of it is inside
*/
int Viewport::RectInViewport(TUI_rectStruct r)
{
	TUI_rectStruct v = TUI_Rect(posX, posY, width / fZoom, height/ fZoom);	
	bool vOverlay = false;
	bool hOverlay = false;
	bool vInside = false;
	
	// vertical overlay
	if (r.y < v.y)
	{
		if (r.y + r.h >v.y)
			vOverlay = true;
	}
	else if (r.y < v.y + v.h)
	{			
		vOverlay = true;
		if (r.y + r.h < v.y + v.h)
			vInside = true;
	}
		
	// horizontal overlay
	if (r.x < v.x)
	{
		if (r.x + r.w >v.x)
			hOverlay = true;
	}
	else if (r.x < v.x + v.w)
	{
		hOverlay = true;				
		if ((vInside) &&(r.x + r.w < v.x + v.w))
			return VIEWPORT_RECT_INSIDE;
	}
		
	if (vOverlay && hOverlay)
		return VIEWPORT_RECT_OVERLAPS;
	else
		return VIEWPORT_RECT_OUTSIDE;
}


/*
* void Viewport::DrawScreenRect(SDL_Surface *surface, float sX, float sY, float eX, float eY)
* Desc: Draw a rectangle on screen using the given world coordinates
* In:   surface is the drawing canvas
*       sX, sY, eX, eY are the world coordinates
*/
void Viewport::DrawScreenRect(SDL_Surface *surface, float sX, float sY, float eX, float eY)
{
	TUI_pointStruct p1, p2;


	p1 =TUI_Point((sX - posX) * fZoom + left, (sY - posY) * fZoom + top);
	p2 =TUI_Point((eX - posX) * fZoom + left, (eY - posY) * fZoom + top);
	
	p1.x = MIN(p1.x, left + width);
	p1.x = MAX(p1.x, left);
	p1.y = MIN(p1.y, top + height);
	p1.y = MAX(p1.y, top);
	
	p2.x = MIN(p2.x, left + width);
	p2.x = MAX(p2.x, left);
	p2.y = MIN(p2.y, top + height);
	p2.y = MAX(p2.y, top);
	
	TUI_Rect(surface, p1, p2, 0, fxInvertDst, false);
}

/*
* void DrawWaypoint(SDL_Surface *surface, float x, float y)
* Desc: Draws a waypoint marker in the viewport using current zoom levels and 
*				given coordinates
* In:		surface is the drawing surface
*				x,y are the world coordinates of the waypoint
*				size is the size of the waypoint os screen
*/
void Viewport::DrawWaypoint(SDL_Surface *surface, int x, int y, int size)
{
	float r = 0;
	TUI_pointStruct p1, p2;

	p1.x = (float)(x - size);
	p1.y = (float)y;
	p2.x = (float)(x + size);
	p2.y = (float)y;
	TUI_Line(p1, p2, MapColor(clBlue, surface->format), surface, fxOverride);

	p1.x = (float)x;
	p1.y = (float)(y - size);
	p2.x = (float)x;
	p2.y = (float)(y + size);
	TUI_Line(p1, p2, MapColor(clBlue, surface->format), surface, fxOverride);

	p1.x = (float)x;
	p1.y = (float)y;
	r = ((float)size / 3);
	TUI_Circle(surface, p1, r, MapColor(clBlue, surface->format), fxOverride);
}

/*
* void DrawWaypoint(SDL_Surface *surface, float x, float y)
* Desc: Draws an AOD circle in the viewport using current zoom levels and 
*				given coordinates
* In:		surface is the drawing surface
*				x,y are the world coordinates of the AOD
*				r is the radius of the circle
*/
void Viewport::DrawAOD(SDL_Surface *surface, int x, int y, int r){
	TUI_pointStruct p;

	p.x = (float)x;
	p.y = (float)y;

	TUI_Circle(surface, p, (float)r, MapColor(clRed, surface->format), fxOverride);
}